#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Simple classes that handle threaded acquisition of data.
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import threading
import time

class DataAcquisitionHandler():
    """ Base class for a handler which acquires new data and stored it in the supplied dictionary 'dataDict'.
        Derive this class and override the 'acquireData' - methods to fill the 'dataDict'.
    """
    
    def __init__(self, dataDict=None):
        """ The contructor gets a dictionary which will be used to store any acquired data.
        """
        self.setDataDict(dataDict)

    def setDataDict(self, dataDict):
        """ Set the shared dictionary
        """
        self.dataDict = dataDict

    def acquireData(self):
        """ This method is called by a 'DataAcquisitionThread'. Override this method to retrieve data and save the data in 'dataDict'.
        """
        pass

class DataAcquisitionThread(threading.Thread):
    """ A thread class which handles cyclic data acquisition.

        An instance of this class needs a lock - object for controlling access to its 'dataAcquisitionHandler', an event - object for 
        notifying the thread to exit and an object of a class derived from 'DataAcquisitionHandler'.
    """
    def __init__(self, threadSleep=0, shutdownFlag=None, dataLock=None, dataAcquisitionHandler=None):
        """ Contructs a new instance of 'DataAcquisitionThread'.
            
            Parameters:
            
            threadSleep             - milliseconds sleep time for the thread loop.
            shutdownFlag            - a threading.event() object for notifying the thread to exit.
            dataLock                - a threading.Lock() object for controlling access to the 'dataAcquisitionHandler'.
            dataAcquisitionHandler  - an object of a class derived from 'DataAcquisitionHandler'.
        """
        threading.Thread.__init__(self)

        self.threadSleep = threadSleep
        self.shutdownFlag = shutdownFlag
        self.dataLock = dataLock
        self.dataAcquisitionHandler = dataAcquisitionHandler
        
    def run(self):
        """ The thread loop runs until 'shutdownFlag' has been signaled. Sleep for 'threadSleep' milliseconds.
        """
        while not self.shutdownFlag.is_set():
            # acquire lock
            self.dataLock.acquire()
            # call data acquisition handler
            self.dataAcquisitionHandler.acquireData()
            # release lock
            self.dataLock.release()
            self.shutdownFlag.wait(self.threadSleep)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  A HTTP server thread that can be stopped by a supplied threading.event() object.
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import threading
import time
from http.server import HTTPServer

class StoppableHttpServer(HTTPServer):
    def __init__(self, server_address, RequestHandlerClass, shutdownFlag=None):
        self.shutdownFlag = shutdownFlag
        return super().__init__(server_address, RequestHandlerClass)

    def serve_forever(self):
        self.socket.settimeout(0.5)
        while not self.shutdownFlag.is_set():
            self.handle_request()

class StoppableHttpServerThread(threading.Thread):
    def __init__(self, 
                 shutdownFlag=None, 
                 handler=None,
                 serverPort=0):
        threading.Thread.__init__(self)

        self.shutdownFlag = shutdownFlag
        self.stoppableHttpServer = StoppableHttpServer(('', serverPort), handler, shutdownFlag)

    def run(self):
        self.stoppableHttpServer.serve_forever()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

__all__ = ["commandmapping", "dataacquisition", "serviceshutdownhandling", "stoppablehttpserver"]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#  
#  Main loop of an Raspend application
#
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import threading
import time

from .http import RaspendHTTPServerThread
from .utils import serviceshutdownhandling as ServiceShutdownHandling
from .utils import dataacquisition as DataAcquisition
from .utils import commandmapping as CommandMapping

class RaspendApplication():
    """ This class handles the main loop for a raspend based application.
    """

    def __init__(self, port, *args, **kwargs):
        # The server port
        self.__port = port

        # A list holding instances of DataAcquisitionThread 
        self.__daqThreads = list()

        # The dictionary holding user's commands he wants to expose.
        self.__cmdMap = CommandMapping.CommandMap()

        # A shared dictionary for the data acquisition threads and the HTTP server thread.
        self.__dataDict = dict()
        
        # Event used for proper shutting down our threads.
        self.__shutdownFlag = threading.Event()
        
        # A lock object for synchronizing access to data within acquistion handlers and the HTTP request handler.
        self.__dataLock = threading.Lock()
        
    def createDataAcquisitionThread(self, dataAcquisitionHandler, threadSleep=1):
        """ This method creates a new instance of 'DataAcquisition.DataAcquisitionThread'.
            Make sure that the handler you provide is derived from 'DataAcquisition.DataAcquisitionHandler'!
        """
        if not isinstance(dataAcquisitionHandler, DataAcquisition.DataAcquisitionHandler):
            raise TypeError("Your 'DataAcquisitionHandler' must be derived from 'DataAcquisition.DataAcquisitionHandler'!")
        
        dataAcquisitionHandler.setDataDict(self.__dataDict)

        dataThread = DataAcquisition.DataAcquisitionThread(threadSleep, 
                                                           self.__shutdownFlag, 
                                                           self.__dataLock, 
                                                           dataAcquisitionHandler)
        self.__daqThreads.append(dataThread)

        return len(self.__daqThreads)

    def addCommand(self, callbackMethod):
        """ Adds a new command to the command map of your application.
        """
        self.__cmdMap.add(CommandMapping.Command(callbackMethod))

        return len(self.__cmdMap)

    def run(self):
        """ Run the main loop of your application.
        """
        try:
            # Initialize signal handler to be able to have a graceful shutdown.
            ServiceShutdownHandling.initServiceShutdownHandling()

            # The HTTP server thread - our HTTP interface
            httpd = RaspendHTTPServerThread(self.__shutdownFlag, self.__dataLock, self.__dataDict, self.__cmdMap, self.__port)

            # Start our threads.
            httpd.start()

            for daqThread in self.__daqThreads:
                daqThread.start()

            # Keep primary thread or main loop alive.
            while True:
                time.sleep(0.5)

        except ServiceShutdownHandling.ServiceShutdownException:
            # Signal the shutdown flag, so the threads can quit their work.
            self.__shutdownFlag.set()
            # Wait for all thread to end.
            for daqThread in self.__daqThreads:
                daqThread.join()
            httpd.join()

        except Exception as e:
            print ("An unexpected error occured. Error: {}".format(e))

        finally:
            pass

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Encapsulates signal handling for graceful shutdown of an application.
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import signal

class ServiceShutdownException(Exception):
    pass

def serviceShutdownHandler(signum, frame):
    raise ServiceShutdownException

def initServiceShutdownHandling():
    signal.signal(signal.SIGTERM, serviceShutdownHandler)
    signal.signal(signal.SIGINT, serviceShutdownHandler)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  This module contains classes for invoking class methods by their names. 
#  For example, it allows you to make these methods available to the outside 
#  world via a network interface. 
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import json
import inspect

class CallbackFunction():
    """
    This class holds a reference to a method of an object.
    
    Usage: 

    obj = classObject()
    cb  = Callback(obj.anyMethod)

    args = dict()

    args["arg1"] = 1
    args["arg2"] = 2

    cb.invoke(args)

    """

    def __init__(self, callbackFunction):
        self.__function = callbackFunction
        self.__name = callbackFunction.__name__
        self.__signature = inspect.signature(callbackFunction)
        self.__objName = self.__findObjectName()

    def __findObjectName(self):
        """
        Determines the variable name of the object whose method was supplied to this instance of class 'CallbackFunction'.
        """
        objToFind = self.__function.__self__
        currStack = inspect.stack()
        for stackFrame in currStack[2:]:
            # Make a copy of the keys to prevent runtime errors because of changes to the dictionary during iteration.
            frameLocalVarNames = list(stackFrame.frame.f_locals)
            for varName in frameLocalVarNames:
                obj = stackFrame.frame.f_locals[varName]
                if obj == objToFind:
                    return varName
        return ""

    def getName(self):
        """
        Returns the name of this instance of class 'CallbackFunction', which is the variable name of the object whose method
        was supplied to this instance and the name of the method itself divided by a period.
        """
        if len(self.__objName):
            return self.__objName + "." + self.__name
        else:
            # We should never come here!
            return self.__function.__qualname__

    def hasParameter(self, parameterName):
        """
        Checks if the signature of the supplied callback method contains 'parameterName'.
        """
        return parameterName in self.__signature.parameters.keys()

    def getParameters(self):
        """
        Returns a dictionary containing the arguments of the supplied callback method.
        """
        return self.__signature.parameters;

    def invoke(self, args):
        """
        Check args and invoke callback method.
        """
        if not type(args) is dict:
            raise TypeError("Arguments need to be passed in a dictionary!")
        for key in args.keys():
            if not self.hasParameter(key):
                raise KeyError("No argument '{0}' found!".format(key))
        return self.__function(**args)

class Command():
    """
    This class wraps a 'CallbackFunction' object.
    """

    def __init__(self, callback):
        self.__callback = CallbackFunction(callback)
        self.__name = self.__callback.getName()

    def getName(self):
        """
        Returns the full name of callback wrapped by this instance.
        """
        return self.__name

    def describe(self, asJSON):
        """
        Returns either a JSON string describing this instance or a dictionary.
        """
        d = dict()
        d["Command"] = dict()
        d["Command"]["Name"] = self.__name
        d["Command"]["Args"] = dict()
        params = self.__callback.getParameters()
        for p in params:
            param = params[p]
            d["Command"]["Args"][param.name] = param.annotation if param.annotation != inspect.Parameter.empty else ""
        if asJSON == True:
            return json.dumps(d, ensure_ascii=False)
        else:
            return d

    def execute(self, args):
        """
        Accepts a dictionary containing the arguments for the callback function. 'key' is an argument's name and 'value' is the value.
        """
        return self.__callback.invoke(args)

class CommandMap(dict):
    """
    A dictionary for holding instances of class 'Command'.
    """

    def __init__(self):
        return super().__init__()

    def add(self, Command):
        """
        Adds 'Command' by its name to the dictionary.
        """
        self[Command.getName()] = Command



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  The HTTP request handling interface for raspend.
#  'raspend' stands for RaspberryPi Backend.
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

from functools import partial
from http.server import BaseHTTPRequestHandler
import json
import urllib

from .utils import stoppablehttpserver


class RaspendHttpRequestHandler(BaseHTTPRequestHandler):
    """ This class handles all incoming request to the raspend server
    """
    def __init__(self, dataLock, dataDict, commandMap, *args, **kwargs):
        self.dataLock = dataLock
        self.dataDict = dataDict
        self.commandMap = commandMap
        return super().__init__(*args, **kwargs)

    def do_OPTIONS(self):
        """ Handle browsers preflight-request to allow POST.
        """
        self.send_response(200)
        self.send_header('Access-Control-Allow-Origin', '*')                
        self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        self.send_header("Access-Control-Allow-Headers", "X-Requested-With, Content-Type") 
        self.end_headers()
        return

    def verifyContentType(self, desiredType, desiredCharset):
        """ Check if 'Content-Type' contains what we want.
        """
        content_type = self.headers.get('Content-Type')
        
        if content_type == None:
            return False
        
        parts = content_type.split(";")
        mediatype = parts[0]
        charset = parts[1] if len(parts) > 1 else ""
        
        isDesiredMediaType = True if mediatype.lower() == desiredType.lower() else False
        isDesiredCharset   = True if desiredCharset.lower() in charset.lower() else False
        
        return isDesiredMediaType and isDesiredCharset

    def verifyPayload(self, payload):
        """ Verifies the format of the payload sent via HTTP POST. 
            As we are dealing with commands that can be invoked via POST, we need to make sure that payload contains what we expect.
        """
        if not type(payload) is dict:
            return False
        if not "Command" in payload.keys():
            return False
        if not "Name" in payload["Command"].keys():
            return False
        return True

    def do_POST(self):
        """ Handle HTTP POST request. Currently this only handles payloads containing a call to a command.
        
            The payload has the following format:
        
            A JavaScript example:
        
            let payLoad = {
                "Command" : {
                    "Name" : "testCmd.methodToCall",
                    "Args" : {
                        "arg1" : 5,
                        "arg2" : 23
                    }
                }
           };
        
           'args' is optional
        
           The payload will be enhanced with the result of the method invocation and returned via JSON to the caller. 
           So you have the opportunity to add any information to the payload e.g. an id of an front-end element connected 
           to the respective command.
        """
        if self.commandMap == None:
            self.send_error(501, "No commands available.")
            return

        if self.verifyContentType("application/json", "utf-8") == False:
            self.send_error(415, "Invalid media type! Expecting 'application/json' using character set 'utf-8'")
            return

        content_length = int(self.headers.get('Content-Length', 0))

        try:
            message_body = self.rfile.read(content_length)
        except BlockingIOError:
            self.send_error(500, "Could not read message body!")
            return

        try:
            payload = json.loads(message_body.decode('UTF-8'))
        except json.JSONDecodeError:
            self.send_error(500, "Encountered a JSON decoding error");
            return

        if self.verifyPayload(payload) == False:
            self.send_error(400, "Unable to verify payload!")
            return

        # Find command handler by its name and execute it with args if given
        
        cmd = self.commandMap.get(payload["Command"]["Name"])

        if cmd == None:
            self.send_error(404, "Command '{0}' not found!".format(payload["Command"]["Name"]))
            return

        args = dict()

        if "Args" in payload["Command"]:
            args = payload["Command"]["Args"]

        result = ""

        try:
            result = cmd.execute(args)
        except Exception as e:
            self.send_error(500, "An unexpected error occured during execution of '{0}'! Exception: {1}".format(payload["Command"]["Name"], e))
            return

        payload["Command"]["Result"] = result

        self.send_response(200)
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Content-type', 'application/json; charset=utf-8')
        self.end_headers()

        self.wfile.write(bytes(json.dumps(payload, ensure_ascii=False), encoding = "UTF-8"))

        return

    def onGetRootDataPath(self):
        """ Called when / is requested and dumps the given 'dataDict' completly.
        """
        self.dataLock.acquire()
        strJsonResponse = json.dumps(self.dataDict, ensure_ascii=False)
        self.dataLock.release()
        return strJsonResponse
                
    def onGetDetailedDataPath(self):
        """ Called when a more detailed path is requested. This allows you to request sub-elements of 'dataDict'.
        """
        pathParts = self.path.split('/')
        self.dataLock.acquire()
        data = self.dataDict
        for part in pathParts[1:]:
            if type(data) is dict and part in data.keys():
                data = data[part]
        strJsonResponse = json.dumps(data, ensure_ascii=False)
        self.dataLock.release()
        return strJsonResponse

    def onGetCmds(self):
        """ Respond with a list of all known commands. Format is a JSON string.
        """
        strJsonResponse = ""
        self.dataLock.acquire()

        cmds = dict()
        cmds["Commands"] = []
        for cmd in self.commandMap.values():
            cmds["Commands"].append(cmd.describe(False))

        strJsonResponse = json.dumps(cmds, ensure_ascii=False)

        self.dataLock.release()
        return strJsonResponse

    def onGetCmd(self, queryParams):
        """ Call a command via HTTP GET. The response will be the command's return value as a JSON string.
        """
        strJsonResponse = ""

        cmdName = queryParams["name"][0]
        cmdArgs = dict()

        cmd = self.commandMap.get(cmdName)

        if cmd == None:
            self.send_error(404, ("Command '{0}' not found!".format(cmdName)))
            return

        del (queryParams["name"])

        for k,v in queryParams.items():
            cmdArgs[k] = v[0]

        result = ""

        try:
            result = cmd.execute(cmdArgs)
        except Exception as e:
            self.send_error(500, "An unexpected error occured during execution of '{0}'! Exception: {1}".format(cmdName, e))
            return

        strJsonResponse = json.dumps(result, ensure_ascii=False)
            
        try:
            self.send_response(200)
            self.send_header('Content-type', 'application/json; charset=utf-8')
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            self.wfile.write(bytes(strJsonResponse, 'utf-8'))
        except OSError:
            self.send_error(500)
        except BlockingIOError:
            self.send_error(500)
        except Exception as e:
            self.send_error(500, "An unexpected error occured during execution of '{0}'! Exception: {1}".format(cmdName, e))
        return

    def do_GET(self):
        """ Handle HTTP GET request

            '/data'     : returns the whole 'dataDict' as JSON string
            '/data/key' : returns sub-element of 'dataDict' as JSON string
            '/cmds'     : returns the list of available commands
        """
        urlComponents = urllib.parse.urlparse(self.path)
        queryParams = urllib.parse.parse_qs(urlComponents.query)

        strJsonResponse = ""

        if urlComponents.path.lower() == "/cmds":
            if self.commandMap == None or len(self.commandMap) == 0:
                self.send_error(501, "No commands available")
                return
            else:
                strJsonResponse = self.onGetCmds()
        elif urlComponents.path.lower() == "/cmd":
            if self.commandMap == None or len(self.commandMap) == 0:
                self.send_error(501, "No commands available")
                return
            else:
                return self.onGetCmd(queryParams)
        elif urlComponents.path.lower() == "/data" and self.dataDict != None:
            strJsonResponse = self.onGetRootDataPath()
        elif urlComponents.path.startswith("/data/")  and self.dataDict != None:
            strJsonResponse = self.onGetDetailedDataPath()
        else:
            self.send_error(404)
            return
        
        try:
            self.send_response(200)
            self.send_header('Content-type', 'application/json; charset=utf-8')
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            self.wfile.write(bytes(strJsonResponse, 'utf-8'))
        except OSError:
            self.send_error(500)
        except BlockingIOError:
            self.send_error(500)

class RaspendHTTPServerThread(stoppablehttpserver.StoppableHttpServerThread):
    """ The raspend server thread using 'RaspendHttpRequestHandler' for request handling
    """
    def __init__(self, shutdownFlag=None, dataLock=None, dataDict=None, commandMap=None, serverPort=0):
        handler = partial(RaspendHttpRequestHandler, dataLock, dataDict, commandMap)
        return super().__init__(shutdownFlag=shutdownFlag, handler=handler, serverPort=serverPort)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#  
#  License: MIT
#  
#  Copyright (c) 2019 Joerg Beckers

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raspend",
    version="1.2.0",
    author="Joerg Beckers",
    author_email="pypi@jobe-software.de",
    description="A small and easy to use HTTP backend framework for the Raspberry Pi which is ideal for small to medium-sized home automation projects.",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jobe3774/raspend.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Home Automation",
    ],
    python_requires='>=3.5',
)